package lsedit;

// Needs to be parent class of EditModeHandler

import java.awt.Container;
import java.awt.Cursor;
import java.awt.Event;
import java.awt.event.MouseEvent;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;

// The Edit and View ModeHandler are toggled by typing 'V'

public class ViewModeHandler extends LandscapeModeHandler /* extends Object */
{
	protected final static int MENU_BASE				= 2000;

	protected final static double SMALL_SCALE_UP		= 1.2; 
	protected final static double SMALL_SCALE_DOWN		= 0.8; 
	protected final static String SMALL_SCALE_STRING	= "20%";

	// Supported modes

	protected MoveModeHandler	  m_moveHandler;
	protected ResizeModeHandler	  m_resizeHandler;
	protected NavigateModeHandler m_navHandler;
	protected GroupModeHandler	  m_groupingHandler;
	protected NewEdgeModeHandler  m_newEdgeModeHandler;

	private  LandscapeModeHandler m_handler = null;
	private	 long			      m_lastUp;
	private  boolean		 	  m_mouseIsDown = false;

	// Supported popups

	protected JPopupMenu		  m_rootEntityPopup;
	protected JPopupMenu		  m_entityPopup;
	protected JPopupMenu		  m_relationPopup;

	protected JPopupMenu buildRootPopup()
	{
		JPopupMenu m;
		JMenu	   m1;
		
		m = new JPopupMenu("Root options");

		m1 = new JMenu("Draw");
		Do.fontMenuItem(m1, m_ls);
		m.add(m1);

		m1 = new JMenu("Move");
		Do.groupAllMenuItem(m1, m_ls);
		m.addSeparator();
		Do.scaleMenuItem(m1, m_ls);
		m.add(m1);
		return(m);
	}
	
	protected JPopupMenu buildEntityPopup()
	{
		JPopupMenu m;
		
		m = new JPopupMenu("Entity options");

		Do.queryMenuItem(m, m_ls);
		m.addSeparator();
		Do.hideMenuItem(m, m_ls);
		m.addSeparator();
		Do.groupAllMenuItem(m, m_ls);
		m.addSeparator();
		Do.scaleMenuItem(m, m_ls);
		return(m);
	}
	
	protected JPopupMenu buildRelationPopup()
	{
		JPopupMenu m;
		
		m = new JPopupMenu("Edge options");
		Do.navigateEdgeMenu(m, m_ls);
		return(m);
	}
	
	// -----------------
	// Protected methods
	// -----------------

	public void processSelectList(Vector edges, int key) 
	{
		Enumeration			en;
		RelationInstance	edge;
		EntityInstance		drawSrc, drawDst;
		Diagram				diagram = m_ls.getDiagram();

		ResultBox resultBox = m_ls.getResultBox();
		resultBox.setResultTitle("SELECTED RELATIONS:");

		for (en = edges.elements(); en.hasMoreElements(); ) {
			edge    = (RelationInstance) en.nextElement();
			drawSrc = edge.getDrawSrc();
			drawDst = edge.getDrawDst();
			switch (key) {
			case Do.EDGE_OPEN_LOW:
				drawSrc.orAllContainedSrcRelationsDstUnder(drawDst);
				break;
			case Do.EDGE_OPEN_SRC:
				drawDst.orDstRelationsSrcUnder(drawDst);
				break;
			case Do.EDGE_OPEN_DST:
				drawSrc.orSrcRelationsDstUnder(drawDst);
				break;
		}	}

		for (en = edges.elements(); en.hasMoreElements(); ) {
			edge    = (RelationInstance) en.nextElement();
			drawSrc = edge.getDrawSrc();
			drawDst = edge.getDrawDst();
			switch (key) {
			case Do.EDGE_OPEN_LOW:
				drawSrc.addSrcDstRelations(RelationInstance.ADD_RESULTS_MARK, resultBox);
				break;
			case Do.EDGE_OPEN_SRC:
				drawDst.addDstRelations(RelationInstance.ADD_RESULTS_MARK, resultBox);
				break;
			case Do.EDGE_OPEN_DST:
				drawSrc.addSrcRelations(RelationInstance.ADD_RESULTS_MARK, resultBox);
				break;
			default:
				resultBox.addRelation(edge);
		}	}

//		resultBox.activate();
		resultBox.done();
		diagram.selectEdges(edges);
	}

	protected void selectEntity(EntityInstance e) 
	{
		Diagram	diagram = m_ls.getDiagram();

		if (e.getGroupFlag()) {
			// Already a member of a group

			if (e != diagram.getKeyEntity()) {
				diagram.setKeyEntity(e);
			}
		} else {
			EntityInstance old_ke = diagram.getKeyEntity();

			if (old_ke != null) {
				diagram.clearGroupFlags();
			}
			diagram.setKeyEntity(e);
		}
	}

	protected void doHandleElision(int key, Object object) 
	{
		Diagram	diagram = m_ls.getDiagram();
		Vector	grp     = diagram.targetEntities(object);

		if (grp == null) {
			return;
		}

		Enumeration en;
		EntityInstance ge;

		for (en = grp.elements(); en.hasMoreElements(); ) {
			ge = (EntityInstance) en.nextElement();

			switch(key)	{
			case Do.SHOW_CONTENTS:
				// Toggle contain elision
				diagram.toggleContainElision(ge);
				break;

			case Do.DST_EDGES:
				diagram.toggleDstElision(ge);
				break;

			case Do.SRC_EDGES:
				diagram.toggleSrcElision(ge);
				break;

			case Do.ENTERING_EDGES:
				if (ge.isOpen()) {
					diagram.toggleEnteringElision(ge);
				}
				break;

			case Do.EXITING_EDGES:
				if (ge.isOpen()) {
					diagram.toggleExitingElision(ge);
				}
				break;
			
			case Do.INTERNAL_EDGES:
				if (ge.isOpen()) {
					diagram.toggleInternalElision(ge);
				}
				break;
			}
		}
		diagram.redrawDiagram();
	}

	protected void handleElision(int key, Object object) 
	{
		String em = "";

		switch(key)	{
		case Do.SHOW_CONTENTS:
			 // Toggle contain elision
			 em = "Containment";
			 doHandleElision(key, object);
			 break;

		case Do.DST_EDGES:
			em = "Target ";
			doHandleElision(key, object);
			break;

		case Do.SRC_EDGES:
			em = "Source "; 
			doHandleElision(key, object);
			break;

		case Do.ENTERING_EDGES:
			em = "Entering";
			doHandleElision(key, object);
			break; 

		case Do.EXITING_EDGES:
			em = "Exiting";
			doHandleElision(key, object);
			break;

		case Do.INTERNAL_EDGES:
			em = "Internal edges";
			doHandleElision(key, object);
			break;
		}

		m_ls.doFeedback(em + " elision toggled for group");
	}

	protected void edgeNavigateSrc(Object object) 
	{
		Diagram			 diagram = m_ls.getDiagram();
		RelationInstance ri      = diagram.targetRelation(object);

		if (ri == null) {
			return;
		}
		navigateTo(ri.getSrc()); 
	}

	protected void edgeNavigateDst(Object object) 
	{
		Diagram			 diagram = m_ls.getDiagram();
		RelationInstance ri      = diagram.targetRelation(object);

		if (ri == null) {
			return;
		}
		navigateTo(ri.getDst()); 
	}

	protected void handleEdgeExpansion(int key, Object object) 
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		visible, src, dst, drawRoot;
		String				msg, msg1;
		Vector				edges;
		Diagram				diagram = m_ls.getDiagram();

		edges = diagram.targetRelations(object);
		if (edges == null) {
			return;
		}
		msg = null;

		switch (key) {
		case Do.EDGE_OPEN_LOW:
			msg1 = "Opened sources and destinations of edges.";
			break;
		case Do.EDGE_OPEN_SRC:
			msg1 = "Opened sources of edges.";
			break;
		case Do.EDGE_OPEN_DST:
			msg1 = "Opened destination of edges.";
			break;
		case Do.EDGE_CLOSE_LOW:
			msg1 = "Closed source and destination of edges";
			break;
		case Do.EDGE_CLOSE_SRC:
			msg1 = "Closed source of edges";
			break;
		default:
			msg1 = "Closed source and destination of edges";
			break;
		}

		drawRoot = diagram.getDrawRoot();

		for (en = edges.elements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();

			switch (key) {
			case Do.EDGE_OPEN_SRC:
			case Do.EDGE_OPEN_LOW:
				visible = ri.getDrawSrc();
				src     = ri.getSrc();
				for (;;) {
					src = src.getContainedBy();
					if (src == null) {
						break;
					}

					if (src.isDstRelationElided(Diagram.CONTAIN_ID)) {
//						System.out.println("Opening " + src);
						src.clearContainElision();
						msg = msg1;
					}
					if (src == visible) {
						break;
				}	}
				break;
			case Do.EDGE_CLOSE_LOW:
			case Do.EDGE_CLOSE_SRC:
				dst = ri.getDst();
				src = ri.getSrc();
				for (;;) {
					src = src.getContainedBy();
					if (src == null) {
						break;
					}
					if (!drawRoot.hasDescendant(src)) {
						break;
					}
					if (src == dst || src.hasDescendant(dst)) {
						break;
					}
					if (!src.isDstRelationElided(Diagram.CONTAIN_ID)) {
						src.setContainElision();
						msg = msg1;
				}	}
				break;
			}

			switch (key) {
			case Do.EDGE_OPEN_DST:
			case Do.EDGE_OPEN_LOW:
				visible = ri.getDrawDst();
				dst     = ri.getDst();
				for (;;) {
					dst = dst.getContainedBy();
					if (dst == null) {
						break;
					}
					if (dst.isDstRelationElided(Diagram.CONTAIN_ID)) {
						dst.clearContainElision();
						msg = msg1;
					}
					if (dst == visible) {
						break;
				}	}
				break;
			case Do.EDGE_CLOSE_LOW:
			case Do.EDGE_CLOSE_DST:
				dst = ri.getDst();
				src = ri.getSrc();
				for (;;) {
					dst = dst.getContainedBy();
					if (dst == null) {
						break;
					}
					if (!drawRoot.hasDescendant(dst)) {
						break;
					}
					if (src == dst || dst.hasDescendant(src)) {
						break;
					}
					if (!dst.isDstRelationElided(Diagram.CONTAIN_ID)) {
						dst.setContainElision();
						msg = msg1;
				}	}
				break;
		}	}

		if (msg != null) {
			processSelectList(edges, key);
			m_ls.doFeedback(msg);
			m_ls.redrawDg();
		} else {
		switch (key) {
			case Do.EDGE_OPEN_LOW:
			case Do.EDGE_OPEN_SRC:
			case Do.EDGE_OPEN_DST:
				msg = "No further expansion is possible";
				break;
			default:
				msg = "No further contraction is possible";
			}	
			m_ls.error(msg);
		}
	}


	// -------
	// Queries
	// -------

	protected boolean goodQueryEdge(EntityInstance src, EntityInstance dst, EntityInstance root)
	{
		if (src.hasDescendant(dst) || dst.hasDescendant(src))
			return false;

		if (src.isClientOrSupplier()) { 
			return root.hasDescendant(dst);
		}
		return true;
	}

	protected int addToForwardEntityList(EntityInstance te, EntityInstance e, RelationClass rc, Vector list)
	{
		Enumeration		 en;
		RelationInstance ri;
		Diagram			 diagram = m_ls.getDiagram();
		EntityInstance	 root = diagram.getDrawRoot();
		EntityInstance	 dst, vdst, ne;

		int num = 0;

//		System.out.println("ViewModeHandler.addToForwardEdgeList " + e);
		for (en = e.srcLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
//			System.out.println("ViewModeHandler.addToForwardEdgeList " + ri);
			if (ri.getRelationClass() == rc) {
				dst  = ri.getDst();
				vdst = ri.getDrawDst();

				if (vdst != null && goodQueryEdge(te, dst, root)) {

					ri.setHighlightFlag();
					dst.setRedBoxFlag();

					if (list.indexOf(vdst) < 0) {
						vdst.setRedBoxFlag();
						if (m_ls.getGroupQueryFlag()) {
							vdst.setGroupFlag();
						}
						list.addElement(vdst);
						num++;
					}	
				}
			}
		}
		return num;
	}

	protected int addForwardRelation(EntityInstance e, RelationClass rc, ResultBox resultBox, boolean alwaysShowReln)
	{
		Vector list = new Vector();

		int num = addToForwardEntityList(e, e, rc, list);

		if (num > 0 || alwaysShowReln) {
			resultBox.addForwardRelation(e, rc, list);
		}
		return num;
	}

	protected int showForwardEdges(EntityInstance e, ResultBox resultBox) 
	{
		int num = 0;

		if (!e.isSupplier()) {
			Enumeration en;
			Diagram		diagram = m_ls.getDiagram();

			for (en = diagram.enumRelationClasses(); en.hasMoreElements(); ) {
				RelationClass rc = (RelationClass) en.nextElement();
//				System.out.println("ViewModeHandler.showForwardEdges class " + rc);
				if (rc.isActive()) {
					num += addForwardRelation(e, rc, resultBox, false);
//					System.out.println("ViewModeHandler.showForwardEdges num " + num);
		}	}	}
		return num;
	}

	// Get all of the relationships from visited destinations to earlier sources

	protected void getSrcRelationWithClosure(EntityInstance e, Vector list)
	{
		Enumeration en;
		Diagram		diagram = m_ls.getDiagram();
		EntityInstance root = diagram.getDrawRoot();

		for (en = e.srcRelationElements(); en.hasMoreElements(); ) {
			RelationInstance ri = (RelationInstance) en.nextElement();

			EntityInstance dst = ri.getDst();

			// Using highlight flag to mark visitation

			boolean visited = dst.isMarked(EntityInstance.REDBOX_MARK);

			ri.setHighlightFlag();

			if (list.indexOf(ri) < 0) 
				list.addElement(ri);

			if (!visited) {
				dst.setRedBoxFlag();
				if (m_ls.getGroupQueryFlag())
					dst.setGroupFlag();
		
				getSrcRelationWithClosure(dst, list);
			}
		}
	}

	protected int addRelations(EntityInstance e, RelationClass rc, Vector list, ResultBox resultBox, boolean src)
	{
		int num = 0;

		if (list.size() > 0) {
			Vector	sublist = new Vector();
			Enumeration en;

			for (en = list.elements(); en.hasMoreElements(); ) {
				RelationInstance ri = (RelationInstance) en.nextElement();

				if (ri.getRelationClass() == rc) {
					ri.setHighlightFlag();
					sublist.add(ri);
					// nsb.append("\t" + ri.getSrc().getLabel() + " " + ri.getDst().getLabel() + "\n");
					num++;
			}	}

			if (num > 0) {
				resultBox.addRelations(e, rc, sublist, src);
			}
		}

		return num;
	}

	// Performed when FORWARD_CLOSURE is pressed

	protected int showSrcEdgesWithClosure(EntityInstance e, ResultBox resultBox) 
	{
		Enumeration   en;
		RelationClass rc;
		Diagram		  diagram = m_ls.getDiagram();
		Vector        list    = new Vector();

		int num = 0;

		diagram.clearQueryFlags();
		getSrcRelationWithClosure(e, list);
		diagram.clearQueryFlags();

		for (en = diagram.enumRelationClasses(); en.hasMoreElements(); ) {
			rc = (RelationClass) en.nextElement();

			if (rc.isActive()) {
				num += addRelations(e, rc, list, resultBox, true);
		}	}
		return num;
	}

	// Backtrace

	public int addToBackEntityList(EntityInstance te, EntityInstance e,	RelationClass rc, Vector list)
	{
		Enumeration      en;
		RelationInstance ri;
		EntityInstance   child, src, vsrc;
		Diagram		     diagram = m_ls.getDiagram();

		int num = 0;

		EntityInstance root = diagram.getDrawRoot();

		for (en = e.dstLiftedRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();

			if (ri.getRelationClass() == rc) {
				src = ri.getSrc();
				vsrc = ri.getDrawSrc();

				if (vsrc != null && goodQueryEdge(src, te, root)) {
					ri.setHighlightFlag();
					src.setRedBoxFlag();

					if (list.indexOf(vsrc) < 0) {
						vsrc.setRedBoxFlag();
						if (m_ls.getGroupQueryFlag())
							vsrc.setGroupFlag();
		
						list.addElement(vsrc);
						num++;
		}	}	}	}
		return num;
	}

	public int addBackRelation(EntityInstance e, RelationClass rc, ResultBox resultBox, boolean alwaysShowReln)
	{
		Vector list = new Vector();

		int num = addToBackEntityList(e, e, rc, list);

		if (num > 0 || alwaysShowReln) {
			resultBox.addBackRelation(e, rc, list);
		}
		return num;
	}

	public int showBackEdges(EntityInstance e, ResultBox resultBox) 
	{
		int num = 0;

		if (!e.isClient()) {

			Enumeration   en;
			RelationClass rc;
			Diagram		  diagram = m_ls.getDiagram();

			for (en = diagram.enumRelationClasses(); en.hasMoreElements(); ) {
				rc = (RelationClass) en.nextElement();
				if (rc.isActive()) {
					num += addBackRelation(e, rc, resultBox, false);
		}	}	}	
		return num;
	}


	protected void getDstRelationWithClosure(EntityInstance e, Vector list)
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		src;
		boolean				visited;

		for (en = e.dstRelationElements(); en.hasMoreElements(); ) {
			ri = (RelationInstance) en.nextElement();
			src = ri.getSrc();

			visited = src.isMarked(EntityInstance.REDBOX_MARK);
			ri.setHighlightFlag();
			if (list.indexOf(ri) < 0) {
				list.addElement(ri);
			}

			if (!visited) {
				src.setRedBoxFlag();
				if (m_ls.getGroupQueryFlag()) {
					src.setGroupFlag();
				}
				getDstRelationWithClosure(src, list);
			}
		}
	}

	// Backward closure

	protected int showDstEdgesWithClosure(EntityInstance e, ResultBox resultBox) 
	{
		Enumeration en;
		Vector		list    = new Vector();
		Diagram		diagram = m_ls.getDiagram();

		int num = 0;

		diagram.clearQueryFlags();
		getDstRelationWithClosure(e, list);
		diagram.clearQueryFlags();


		for (en = diagram.enumRelationClasses(); en.hasMoreElements(); ) {
			RelationClass rc = (RelationClass) en.nextElement();

			if (rc.isActive()) {
				num += addRelations(e, rc, list, resultBox, false);
		}	}
		return num;
	}

	protected int showContents(EntityInstance e, ResultBox resultBox) 
	{
		Enumeration en;
		Vector list = new Vector();
		int n = 0;

		for (en = e.getChildren(); en.hasMoreElements(); ) {
			list.add(en.nextElement());
			n++;
		}
		if (n > 0) {
			SortVector.byString(list);
		}
		resultBox.addContents(e, list);
		return n;
	}

	protected int showContentsWithClosure(EntityInstance e, ResultBox resultBox)
	{ 
		Enumeration en;
		int			n;

		n = showContents(e, resultBox);

		for (en = e.getChildren(); en.hasMoreElements(); ) {
			n += showContentsWithClosure((EntityInstance) en.nextElement(), resultBox);
		}
		return n;
	}

	protected int doQueryEntity(EntityInstance e, int query, ResultBox resultBox) 
	{
		switch(query) {
		case Do.FORWARD_QUERY:
			return showForwardEdges(e, resultBox);

		case Do.BACKWARD_QUERY:
			return showBackEdges(e, resultBox);

		case Do.FORWARD_CLOSURE:
			return showSrcEdgesWithClosure(e, resultBox);

	
		case Do.BACKWARD_CLOSURE:
			return showDstEdgesWithClosure(e, resultBox);

		case Do.CONTENTS_QUERY:
			return showContents(e, resultBox);

		case Do.CONTENT_CLOSURE:
			return showContentsWithClosure(e, resultBox);

		default:
			return 0;
		}
	}

	protected void queryEntity(int query, Object object) 
	{
		Enumeration		en;
		ResultBox		resultBox  = m_ls.getResultBox();
		Diagram		    diagram    = m_ls.getDiagram();
		Vector			grp        = diagram.targetEntities(object);
		String			title;

		if (!diagram.getQueryPersistance()) {
			diagram.clearQueryFlags();
		}

		switch(query) {
		case Do.FORWARD_QUERY:
			title = "FORWARD QUERY";
			break;
		case Do.FORWARD_CLOSURE:
			title = "FORWARD QUERY (closure)";
			break;
		case Do.BACKWARD_QUERY:
			title = "BACKWARD QUERY";
			break;
		case Do.BACKWARD_CLOSURE:
			title = "BACKWARD QUERY (closure)";
			break;
		default:
			title = null;
		}

		resultBox.setResultTitle(title);

		if (grp == null || grp.isEmpty()) {
			m_ls.error("No entity or group is selected");
			resultBox.done();
			return;
		}

		int num = 0;

		for (en = grp.elements(); en.hasMoreElements(); ) {
			EntityInstance ge = (EntityInstance) en.nextElement();

			num += doQueryEntity(ge, query, resultBox);
		}

		if (num == 0) {
			resultBox.addText("NO ENTITIES");
		}
		resultBox.validate();
		diagram.clearDrawEdges();
	}

	protected void queryContents(int key) 
	{
		Enumeration		en;
		ResultBox		resultBox  = m_ls.getResultBox();
		Diagram		    diagram    = m_ls.getDiagram();
		Vector			grp        = diagram.getGroup();
		EntityInstance  ge;
		String			title;

		switch (key) {
			case Do.CONTENT_CLOSURE:
				title = "CONTENTS (CLOSURE)";
				break;
			case Do.CONTENTS_QUERY:
				title = "CONTENTS";
				break;
			default:
				title = null;
		}

		resultBox.setResultTitle(title);

		if (grp == null || grp.isEmpty()) {
			doQueryEntity(diagram.getDrawRoot(), key, resultBox);
			resultBox.validate();
			return;
		}

		int num = 0;

		for (en = grp.elements(); en.hasMoreElements(); ) {
			ge = (EntityInstance) en.nextElement();
			num += doQueryEntity(ge, key, resultBox);
		}

		if (num == 0) {
			resultBox.addText("NO ENTITIES");
		}
		resultBox.validate();
	}

	protected void doScaleEntity(EntityInstance ge, int scale, boolean incFlag) 
	{
		switch(scale) {
		case Do.DECREASE_WIDTH:
			ge.scaleX(SMALL_SCALE_DOWN, incFlag);
			break; 

		case Do.INCREASE_WIDTH:
			ge.scaleX(SMALL_SCALE_UP, incFlag);
			break; 

		case Do.DECREASE_HEIGHT:
			ge.scaleY(SMALL_SCALE_DOWN, incFlag);
			break; 

		case Do.INCREASE_HEIGHT:
			ge.scaleY(SMALL_SCALE_UP, incFlag);
			break; 

		case Do.DECREASE_SIZE:
			ge.scale(SMALL_SCALE_DOWN, incFlag);
			break;

		case Do.INCREASE_SIZE:
			ge.scale(SMALL_SCALE_UP, incFlag);
			break;
		}
	}

	protected void scaleEntity(int scale, boolean incFlag) 
	{
		Diagram	diagram = m_ls.getDiagram();
		Vector	grp     = diagram.getGroup();

		if (grp != null) {
			Enumeration en;

			switch (grp.size()) {
			case 0:
				break;
			case 1:
				if (grp.firstElement() == diagram.getDrawRoot()) {
					break;
				}
			default:
				for (en = grp.elements(); en.hasMoreElements(); ) {
					EntityInstance ge = (EntityInstance) en.nextElement();
					doScaleEntity(ge, scale, incFlag);
				}
				return;
			}
		}

		// Do it to root

		switch (scale) {
		case Do.INCREASE_SIZE:
			// Zoom in
			if (diagram.zoom(SMALL_SCALE_UP, SMALL_SCALE_UP)) {
				m_ls.doFeedback( "Zoomed in (" + SMALL_SCALE_STRING + ")");
			}
			return;

		case Do.DECREASE_SIZE:
			// Zoom out
			if (diagram.zoom(SMALL_SCALE_DOWN, SMALL_SCALE_DOWN)) {
				m_ls.doFeedback( "Zoomed out (" + SMALL_SCALE_STRING + ")");
			}
			return;

		case Do.INCREASE_WIDTH:
			if (diagram.zoom(SMALL_SCALE_UP, 1.0)) {
				m_ls.doFeedback("Scaled X direction (" + SMALL_SCALE_STRING + ")");
			}
			return;

		case Do.DECREASE_WIDTH:
			if (diagram.zoom(SMALL_SCALE_DOWN, 1.0)) {
				m_ls.doFeedback("Scaled X direction (-" + SMALL_SCALE_STRING + ")");
			}
			return;

		case Do.INCREASE_HEIGHT:
			if (diagram.zoom(1.0, SMALL_SCALE_UP)) {
				m_ls.doFeedback("Scaled Y direction (" + SMALL_SCALE_STRING + ")");
			}
			return;

		case Do.DECREASE_HEIGHT:
			if (diagram.zoom(1.0, SMALL_SCALE_DOWN)) {
				m_ls.doFeedback("Scaled Y direction (-" + SMALL_SCALE_STRING + ")");
			}
			return;
		}
	}

	protected boolean exclusiveEntityCommand(int key) 
	{
		switch(key)	{
		case Do.DESCEND:
		case Do.MOVE_GROUP_UP:
		case Do.MOVE_GROUP_DOWN:
		case Do.MOVE_GROUP_LEFT:
		case Do.MOVE_GROUP_RIGHT:
			return true;
		}

		return false;
	}

	protected void navigateTo(EntityInstance e) 
	{
		m_ls.followLink(e, false);
	}

	// Overloaded by EditModeHandler

	protected boolean handleCommands(int key, int modifiers, Object object) 
	{
		Diagram	diagram = m_ls.getDiagram();

//		System.out.println("ViewModeHandler:handleCommands(" + key + ", " + ", " + modifiers + ", " + object +")");

		if (modifiers == Event.CTRL_MASK) {
			switch(key)	{
			case Do.GROUP_ALL:
				m_ls.groupAll();
				break;
			default:
				return false;
			}
		} else if ((modifiers & Event.ALT_MASK) != 0) {
			// Alt key set if get here
			switch(key)	{
			case Do.CONTENT_CLOSURE:
				queryContents(key);
				break;
			case Do.EDGE_OPEN_LOW:
			case Do.EDGE_OPEN_SRC:
			case Do.EDGE_OPEN_DST:
			case Do.EDGE_CLOSE_LOW:
			case Do.EDGE_CLOSE_SRC:
			case Do.EDGE_CLOSE_DST:
				handleEdgeExpansion(key, object);
				return true;
			case Do.EDGE_NAVIGATE_SRC:
				edgeNavigateSrc(object);
				return true;
			case Do.EDGE_NAVIGATE_DST:
				edgeNavigateDst(object);
				return true;
			default:
				return false;
			}
		} else if (modifiers == Event.SHIFT_MASK) {
			switch (key) {
			case Do.ASCEND:	// Special key SHIFT still set
			{
				EntityInstance e = diagram.targetEntity(object);
				if (e == null) {
					return false;
				}
				diagram.setDrawRoot(e);
				diagram.navigateToDrawRootParent();
				break;
			}
			default:
				return false;
			}
		} else {
			switch(key)	{
			case Do.CONTENTS_QUERY:
				queryContents(key);
				break;
			case Do.FORWARD_QUERY:
			case Do.FORWARD_CLOSURE:
			case Do.BACKWARD_QUERY:
			case Do.BACKWARD_CLOSURE:		
				queryEntity(key, object);
				break;

			case Do.SHOW_CONTENTS:
			case Do.DST_EDGES:
			case Do.ENTERING_EDGES:
			case Do.SRC_EDGES:
			case Do.EXITING_EDGES:
			case Do.INTERNAL_EDGES:
				handleElision(key, object);
				break;

			case Do.DECREASE_WIDTH:
			case Do.INCREASE_WIDTH:
			case Do.DECREASE_HEIGHT:
			case Do.INCREASE_HEIGHT:
			case Do.DECREASE_SIZE:
			case Do.INCREASE_SIZE:
				m_ls.beginUndoRedo("Scale");
				scaleEntity(key, !m_mouseIsDown);
				m_ls.endUndoRedo();
				break;
			case Do.MOVE_GROUP_UP:
			case Do.MOVE_GROUP_DOWN:
			case Do.MOVE_GROUP_LEFT:
			case Do.MOVE_GROUP_RIGHT:
				m_ls.beginUndoRedo("Move");
				m_groupingHandler.moveGroup(key);
				m_ls.endUndoRedo();
				break;

			case Do.DESCEND:
			{
				EntityInstance e = diagram.targetEntity(object);
				if (e == null) {
					return false;
				}

				EntityInstance drawRoot = diagram.getDrawRoot();

				if (e == drawRoot) {
					m_ls.error("Already in: " + e.getLabel());
				} else if (e == drawRoot.getContainedBy()) {
					// Going up
					navigateTo(e);
				} else if (e.isClientOrSupplier()) {
					if (e.hasChildren()) {
						navigateTo(e);
					} else {
						navigateTo(e.getEnterableParent());
					}
				} else {
					// Going down

					while (e.getContainedBy() != drawRoot && !e.getContainedBy().isEnterable()) {
						e = e.getContainedBy();
					}

					if (e.isEnterable()) {
						navigateTo(e);
					} else {
						m_ls.error("Can't navigate into: " + e.getLabel());
					}
				}
				break;
			}
			default:
				return false;
			}
		} 
		m_ls.redrawDg();
		return true;
	}

	// --------------
	// Public methods
	// --------------

	public ViewModeHandler(LandscapeViewerCore ls) 
	{
		super(ls);

		m_moveHandler        = new MoveModeHandler(this);
		m_resizeHandler      = new ResizeModeHandler(this);
		m_navHandler         = new NavigateModeHandler(this);
		m_groupingHandler    = new GroupModeHandler(this);
	}

	public 	LandscapeModeHandler getSubHandler()
	{
//		System.out.println("ViewModeHandler.getSubHandler");
		return(m_handler);
	}

	public 	void setSubHandler(LandscapeModeHandler handler)
	{
//		System.out.println("ViewModeHandler.setSubHandler from " + m_handler + " to " + handler);
		if (m_handler != null) {
			m_handler.cleanup();
		}
		m_handler = handler;
	}

	public void cleanup() 
	{
//		System.out.println("ViewModeHandler.cleanup");
		setSubHandler(null);
	}

	public void processKey(int key, int modifiers, Object object) 
	{
//		System.out.println("ViewModeHandler:processKey(" + key + ", " + ", " + modifiers + ", " + object +")");

		if ((modifiers & Event.ALT_MASK) != 0) {
			switch (key) {
			case Do.EDGE_OPEN_LOW:
			case Do.EDGE_OPEN_SRC:
			case Do.EDGE_OPEN_DST:
			case Do.EDGE_CLOSE_LOW:
			case Do.EDGE_CLOSE_SRC:
			case Do.EDGE_CLOSE_DST:
				handleEdgeExpansion(key, object);
				return;
			case Do.EDGE_NAVIGATE_SRC:
				edgeNavigateSrc(object);
				return;
			case Do.EDGE_NAVIGATE_DST:
				edgeNavigateDst(object);
				return;
		}	}

		handleCommands(key, modifiers, object);
	}

	protected static void usesObject(Object c, Object object)
	{
		if (c instanceof MyMenuItem) {
			((MyMenuItem) c).setObject(object);
			return;
		}
		if (c instanceof JMenu) {
			JMenu 		c1;
			JMenuItem	item;
			int		i;
			
			c1 = (JMenu) c;
			for (i = c1.getItemCount(); i > 0; ) {
				item = c1.getItem(--i);
				if (item != null) {
					usesObject(item, object);
			}	}
			return;
		}				
		if (c instanceof Container) {
			Container	c1;
			int			i;

			c1 = (Container) c;
			for (i = c1.getComponentCount(); i > 0; ) {
				usesObject(c1.getComponent(--i), object);
			}
			return;
		}
//		System.out.println("ViewModeHandler.usesObject " + c);
	}

		
	public void rightClickEntity(MouseEvent ev, EntityInstance e, int x, int y)
	{
		Diagram		diagram = m_ls.getDiagram();
		JPopupMenu	m;

		if (e == diagram.getRootInstance()) {
			m = m_rootEntityPopup;
			if (m == null) {
				m_rootEntityPopup = m = buildRootPopup();
			}
		} else {
			m = m_entityPopup;
			if (m == null) {
				m_entityPopup = m = buildEntityPopup();
		}	}
	
		usesObject(m, e);
		m_ls.add(m);
		m.show(m_ls.getContentPane(), m_ls.getDiagramX()+x, m_ls.getDiagramY()+y);
	}

	public void rightClickRelation(MouseEvent ev, RelationInstance ri, int x, int y)
	{
		JPopupMenu	m;

		m = m_relationPopup;
		if (m == null) {
			m_relationPopup = m = buildRelationPopup();
		}
		usesObject(m, ri);
		m_ls.add(m);
		m.show(m_ls.getContentPane(), m_ls.getDiagramX()+x, m_ls.getDiagramY()+y);
	}

	public void entityPressed(MouseEvent ev, EntityInstance e, int x, int y) 
	{
//		System.out.println("ViewModeHandler entity mousedown\n");

		if (ev.isMetaDown()) {
			rightClickEntity(ev, e, x, y);
			return;
		}

		m_mouseIsDown     = true;
//		System.out.println("ViewModeHandler.entityPressed " + m_handler);
		if (m_handler != null) {
			m_handler.entityPressed(ev, e, x, y);
			return;
		}
		Diagram	diagram = m_ls.getDiagram();

		if (e == diagram.getDrawRoot()) {
			if (diagram.clearFlags()) {
				diagram.redrawDiagram();
		}	}

		if (ev.isControlDown()) {
			m_navHandler.entityPressed(ev, e, x, y);
			return;
		} 
		if (ev.isShiftDown()) {
			m_groupingHandler.entityPressed(ev, e, x, y);
			return;
		}
		// Determine if bend/IO move, resize, or group move/activate

		m_resizeHandler.entityPressed(ev, e, x, y);
		if (m_handler != null) {
			return;
		}
		m_moveHandler.entityPressed(ev, e, x, y);
		if (m_handler != null) {
			return;
		}
		m_groupingHandler.entityPressed(ev, e, x, y);
	}

	public void entityReleased(MouseEvent ev, EntityInstance e, int x, int y)
	{
		if (m_mouseIsDown) {
			long newUp = ev.getWhen();

//			System.out.println("ViewModeHandler.entityReleased " + m_handler);
			if (newUp - m_lastUp < 300) {
//				System.out.println("Double click");
				m_ls.setCursor(Cursor.DEFAULT_CURSOR);
				mouseDoubleClick(ev, e);
				m_lastUp = 0;
			} else if (m_handler != null) {
				m_handler.entityReleased(ev, e, x, y);
			}
			m_lastUp    = newUp;
			m_mouseIsDown = false;
	}	}

	public void relationPressed(MouseEvent ev, RelationInstance ri, int x, int y) 
	{

//		System.out.println("ViewModeHandler edge mousedown\n");

		if (ev.isMetaDown()) {
			rightClickRelation(ev, ri, x, y);
			return;
		}
		Diagram	diagram = m_ls.getDiagram();
		m_mouseIsDown   = true;

		if (m_handler != null) {
//			System.out.println("ViewModeHandler.relationPressed " + m_handler);
			m_handler.relationPressed(ev, ri, x, y);
			return;
		} 
		diagram.clearFlags();

		if (!ev.isControlDown()) {
			if (ev.isShiftDown()) {
				m_groupingHandler.relationPressed(ev, ri, x, y);
				return;
			}
			// Determine if edge point move
			m_moveHandler.relationPressed(ev, ri, x, y);
			if (m_handler == null) {
				m_groupingHandler.relationPressed(ev, ri, x, y);
		}	}	
	}

	public void relationReleased(MouseEvent ev, RelationInstance ri, int x, int y)
	{
		if (m_mouseIsDown) {
			long newUp = ev.getWhen();

			if (newUp - m_lastUp < 300) {

//				System.out.println("Double click");
				m_ls.setCursor(Cursor.DEFAULT_CURSOR);
				mouseDoubleClick(ev, ri);
				m_lastUp = 0;
			} else {
				m_lastUp = newUp;
//				System.out.println("ViewModeHandler.relationReleased " + m_handler);
				if (m_handler != null) {
					m_handler.relationReleased(ev, ri, x, y);
			}	}
		}
		m_mouseIsDown = false;
	}
	
	public void mouseDoubleClick(MouseEvent ev, Object object) 
	{
		m_ls.processKey(Do.DESCEND, (ev.getModifiers() & (Event.SHIFT_MASK|Event.ALT_MASK|Event.CTRL_MASK|Event.META_MASK)), object);
	}

	public void movedOverThing(MouseEvent ev, Object thing, int x, int y) 
	{
//		System.out.println("ViewModeHandler.moveOverThing");
		if (m_handler != null) {
			m_handler.movedOverThing(ev, thing, x, y);
			return;
		}
		m_resizeHandler.movedOverThing(ev, thing, x, y);
		if (m_handler == null) {
			m_moveHandler.movedOverThing(ev, thing, x, y);
	}	}

	public void entityDragged(MouseEvent ev, EntityInstance e, int x, int y) 
	{
//		System.out.println("ViewModeHandler.entityDragged " + m_handler);
		if (m_handler != null) {
			m_handler.entityDragged(ev, e, x, y);
		}
	}

	public void relationDragged(MouseEvent ev, RelationInstance ri, int x, int y) 
	{
//		System.out.println("ViewModeHandler.relationDragged " + m_handler);
		if (m_handler != null) {
			m_handler.relationDragged(ev, ri, x, y);
		}
	}
}
